#  Copyright 2020 Google LLC
#  Copyright 2020 EPAM Systems
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Auth and refresh token utility functions."""

import json
import logging
import os
from abc import ABC, abstractmethod
from functools import partial
from http import HTTPStatus
from typing import Any, Callable, Union

import requests
from libs.exceptions import TokenRefresherNotPresentError
from osdu_api.libs.auth.authorization import TokenRefresher as OSDUAPITokenRefresher
from providers import credentials
from providers.types import BaseCredentials
from tenacity import retry, stop_after_attempt

logger = logging.getLogger(__name__)

RETRIES = 3


class TokenRefresher(ABC):
    """Abstract base class for token refreshers."""

    @abstractmethod
    def refresh_token(self) -> str:
        """Refresh auth token.

        :return: refreshed token
        :rtype: str
        """
        pass

    @property
    @abstractmethod
    def access_token(self) -> str:
        """Auth access token.

        :return: token string
        :rtype: str
        """
        pass

    @property
    @abstractmethod
    def authorization_header(self) -> dict:
        """Authorization header. Must return  authorization header for
        updating headers dict.
        E.g. return {"Authorization": "Bearer <access_token>"}

        :return: A dictionary with authorization header updated
        :rtype: dict
        """
        pass


class AirflowTokenRefresher(TokenRefresher, OSDUAPITokenRefresher):
    """Simple wrapper for credentials to be used in refresh_token decorator within Airflow."""

    def __init__(self, creds: BaseCredentials = None):
        self._credentials = creds or credentials.get_credentials()
        from airflow.models import Variable
        self.airflow_variables = Variable
        self._access_token = None

    @retry(stop=stop_after_attempt(RETRIES))
    def refresh_token(self) -> str:
        """Refresh the token and cache token using airflow variables.

        :return: The refreshed token
        :rtype: str
        """
        self._credentials.refresh_token()
        self.airflow_variables.set("access_token", self._credentials.access_token)
        self._access_token = self._credentials.access_token
        return self._access_token

    @property
    def access_token(self) -> str:
        """The access token.

        :return: The access token
        :rtype: str
        """
        if not self._access_token:
            try:
                self._access_token = self.airflow_variables.get("access_token")
            except KeyError:
                self.refresh_token()
        return self._access_token

    @property
    def authorization_header(self) -> dict:
        """Authorization header with bearer token.

        :return: Auth header as dict
        :rtype: dict
        """
        return {"Authorization": f"Bearer {self.access_token}"}


def make_callable_request(obj: Union[object, None], request_function: Callable, headers: dict,
                          *args, **kwargs) -> Callable:
    """Create send_request_with_auth function.

    :param obj: If wrapping a method the obj passed as first argument (self)
    :type obj: Union[object, None]
    :param request_function: The function to be build
    :type request_function: Callable
    :param headers: The request headers
    :type headers: dict
    :return: A partial callable
    :rtype: Callable
    """
    if obj:  # if wrapped function is an object's method
        callable_request = partial(request_function, obj, headers, *args, **kwargs)
    else:
        callable_request = partial(request_function, headers, *args, **kwargs)
    return callable_request


def _validate_headers_type(headers: dict):
    if not isinstance(headers, dict):
        logger.error(f"Got headers {headers}")
        raise TypeError("Request's headers type expected to be 'dict'")


def _validate_response_type(response: requests.Response, request_function: Callable):
    if not isinstance(response, requests.Response):
        logger.error(f"Function or method {request_function}"
                     f" must return values of type 'requests.Response'. "
                     f"Got {type(response)} instead")
        raise TypeError


def _validate_token_refresher_type(token_refresher: TokenRefresher):
    if not isinstance(token_refresher, TokenRefresher):
        raise TypeError(
            f"Token refresher must be of type {TokenRefresher}. Got {type(token_refresher)}"
        )


def _get_object_token_refresher(
    token_refresher: TokenRefresher,
    obj: object = None
) -> TokenRefresher:
    """Get token refresher passed into decorator or specified in object's as
    'token_refresher' property.

    :param token_refresher: A token refresher instance
    :type token_refresher: TokenRefresher
    :param obj: The holder object of the decorated method, defaults to None
    :type obj: object, optional
    :raises TokenRefresherNotPresentError: When not found
    :return: The token refresher
    :rtype: TokenRefresher
    """
    if token_refresher is None and obj:
        try:
            obj.__getattribute__("token_refresher")
        except AttributeError:
            raise TokenRefresherNotPresentError("Token refresher must be passed into decorator or "
                                                "set as object's 'refresh_token' attribute.")
        else:
            token_refresher = obj.token_refresher
    return token_refresher


def send_request_with_auth_header(token_refresher: TokenRefresher, *args,
                                  **kwargs) -> requests.Response:
    """Send request with authorization token. If response status is in
    HTTPStatus.UNAUTHORIZED or HTTPStatus.FORBIDDEN, then refreshes token
    and sends request once again.

    :param token_refresher: The token refresher instance
    :type token_refresher: TokenRefresher
    :raises e: Re-raises any requests.HTTPError
    :return: The server response
    :rtype: requests.Response
    """
    obj = kwargs.pop("obj", None)
    request_function = kwargs.pop("request_function")
    headers = kwargs.pop("headers")
    _validate_headers_type(headers)
    headers.update(token_refresher.authorization_header)

    send_request_with_auth = make_callable_request(obj, request_function, headers, *args, **kwargs)
    response = send_request_with_auth()
    _validate_response_type(response, request_function)

    if not response.ok:
        if response.status_code in (HTTPStatus.UNAUTHORIZED, HTTPStatus.FORBIDDEN):
            token_refresher.refresh_token()
            headers.update(token_refresher.authorization_header)
            send_request_with_auth = make_callable_request(obj,
                                                           request_function,
                                                           headers,
                                                           *args, **kwargs)
            response = send_request_with_auth()
        try:
            response.raise_for_status()
        except requests.HTTPError as e:
            logger.error(f"{response.text}")
            raise e
    return response


def refresh_token(token_refresher: TokenRefresher = None) -> Callable:
    """Wrap a request function and check response. If response's error status
    code is about Authorization, refresh token and invoke this function once
    again.
    Expects function.
    If response is not ok and not about Authorization, then raises HTTPError
    request_func(header: dict, *args, **kwargs) -> requests.Response
    Or method:
    request_method(self, header: dict, *args, **kwargs) -> requests.Response

    :param token_refresher: [description], defaults to None
    :type token_refresher: TokenRefresher, optional
    :return: [description]
    :rtype: Callable
    """

    def refresh_token_wrapper(request_function: Callable) -> Callable:
        is_method = len(request_function.__qualname__.split(".")) > 1
        if is_method:
            def _wrapper(obj: object, headers: dict, *args, **kwargs) -> requests.Response:
                _token_refresher = _get_object_token_refresher(token_refresher, obj)
                _validate_token_refresher_type(_token_refresher)
                return send_request_with_auth_header(_token_refresher,
                                                     request_function=request_function,
                                                     obj=obj,
                                                     headers=headers,
                                                     *args,
                                                     **kwargs)
        else:
            def _wrapper(headers: dict, *args, **kwargs) -> requests.Response:
                _validate_token_refresher_type(token_refresher)
                return send_request_with_auth_header(token_refresher,
                                                     request_function=request_function,
                                                     headers=headers,
                                                     *args, **kwargs)
        return _wrapper

    return refresh_token_wrapper
