#  Copyright 2021 Google LLC
#  Copyright 2021 EPAM Systems
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import copy
import logging
from typing import List

from libs.exceptions import EmptyManifestError


logger = logging.getLogger()


class ManifestTraversal(object):
    """Class to traverse manifest and extract all manifest records"""

    def __init__(self, dagrun_conf: dict, manifest_schema: dict):
        self.data_object = copy.deepcopy(dagrun_conf)
        self.manifest_schema = manifest_schema
        self.manifest_info = []

    def _populate_manifest_entity(self, entity: dict, schema: str):
        """
        Populate manifest entity for future processing

        :param entity: manifest entity instance (for future processing)
        :param schema: corresponding generic schema (for future schema validation)
        :return:
        """
        extracted_schema = schema.split("/")[-1]
        logger.debug(f"Extracted schema kind: {extracted_schema}")
        return {
            "schema": extracted_schema,
            "entity": entity
        }

    def _traverse_list(self, manifest_entities: List[dict], property_name: str, manifest_schema_part: dict):
        """
        Traverse list of entities and returned populated list of entities
        """
        entities = []
        for manifest_entity in manifest_entities:
            entities.append(
                self._populate_manifest_entity(manifest_entity,
                                               manifest_schema_part[property_name]["items"]["$ref"]))
        return entities

    def traverse_manifest(self) -> List[dict]:
        """
        Traverse manifest structure and return the list of manifest records.

        :return: list of records
        """
        if "manifest" not in self.data_object:
            raise EmptyManifestError

        manifest_entities = []
        manifest_file = self.data_object["manifest"]
        if manifest_file.get("ReferenceData"):
            manifest_entities.extend(self._traverse_list(manifest_file["ReferenceData"],
                                                         "ReferenceData",
                                                         self.manifest_schema["properties"]))
        if manifest_file.get("MasterData"):
            manifest_entities.extend(self._traverse_list(manifest_file["MasterData"],
                                                         "MasterData",
                                                         self.manifest_schema["properties"]))
        if manifest_file.get("Data"):
            if manifest_file["Data"].get("WorkProduct"):
                manifest_entities.append(self._populate_manifest_entity(
                        manifest_file["Data"]["WorkProduct"],
                        self.manifest_schema["properties"]["Data"]["properties"]["WorkProduct"]["$ref"]))
            if manifest_file["Data"].get("WorkProductComponents"):
                manifest_entities.extend(self._traverse_list(manifest_file["Data"]["WorkProductComponents"],
                                                             "WorkProductComponents",
                                                             self.manifest_schema["properties"]["Data"]["properties"]))
            if manifest_file["Data"].get("Datasets"):
                manifest_entities.extend(self._traverse_list(manifest_file["Data"]["Datasets"],
                                                             "Datasets",
                                                             self.manifest_schema["properties"]["Data"]["properties"]))
        return manifest_entities
