#  Copyright 2020 Google LLC
#  Copyright 2020 EPAM Systems
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""R2 Process Manifest operator."""

import logging
from airflow.utils import apply_defaults
from airflow.models import BaseOperator, Variable
from libs.context import Context
from libs.source_file_check import SourceFileChecker
from libs.handle_file import FileHandler
from libs.refresh_token import AirflowTokenRefresher
from libs.process_manifest_r3 import ManifestProcessor
from libs.traverse_manifest import ManifestTraversal
from libs.validate_schema import SchemaValidator

logger = logging.getLogger()


class ProcessManifestOperatorR3(BaseOperator):
    """Operator to process manifest R3."""

    ui_color = '#dad5ff'
    ui_fgcolor = '#000000'

    @apply_defaults
    def __init__(self, *args, **kwargs):
        """Init base operator and obtain base urls from Airflow Variables."""
        super().__init__(*args, **kwargs)
        self.schema_service_url = Variable.get('schema_service_url')
        self.storage_url = Variable.get('storage_url')
        self.file_service_url = Variable.get('file_service_url')

    def execute(self, context: dict):
        """Execute manifest validation then process it.

        Execution steps:
        1) initialize schema validator
        2) validate manifest file against common schema
        3) traverse manifest file and extract manifest entities
        4) validate extracted manifest entities
        5) process valid manifest entities

        :param context: Airflow context
        :type context: dict
        """

        execution_context = context["dag_run"].conf["execution_context"]
        payload_context = Context.populate(execution_context)
        token_refresher = AirflowTokenRefresher()
        file_handler = FileHandler(self.file_service_url, token_refresher, payload_context)
        source_file_checker = SourceFileChecker()

        validator = SchemaValidator(
            self.schema_service_url,
            token_refresher,
            payload_context
        )
        manifest_schema = validator.validate_common_schema(execution_context)
        traversal = ManifestTraversal(execution_context, manifest_schema)
        manifest_entities = traversal.traverse_manifest()
        logger.debug(f"entities count: {len(manifest_entities)}")
        valid_manifest_entities = validator.validate_manifest(manifest_entities)
        logger.debug(f"valid entities count: {len(valid_manifest_entities)}")
        manifest_processor = ManifestProcessor(
            self.storage_url,
            valid_manifest_entities,
            file_handler,
            source_file_checker,
            token_refresher,
            payload_context,
        )
        record_ids = manifest_processor.process_manifest()
        context["ti"].xcom_push(key="record_ids", value=record_ids)
